C***********************************************************************
C    Module:  dsopt.f
C 
C    Copyright (C) 2008 Mark Drela 
C 
C    This program is free software; you can redistribute it and/or modify
C    it under the terms of the GNU General Public License as published by
C    the Free Software Foundation; either version 2 of the License, or
C    (at your option) any later version.
C
C    This program is distributed in the hope that it will be useful,
C    but WITHOUT ANY WARRANTY; without even the implied warranty of
C    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
C    GNU General Public License for more details.
C
C    You should have received a copy of the GNU General Public License
C    along with this program; if not, write to the Free Software
C    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
C***********************************************************************

      program dsopt
c-------------------------------------------------------------------------
c     Calculates aero quantities over a Dynamic Soaring orbit 
c     of prescribed shape.  The orbit is computed so as to be 
c     in equilibrium, with the same initial and final states.
c
c     Optionally optimizes the airplane mass and orbit parameters
c     to get the maximum DS speed.
c
c     Usage:
c
c       % dsopt inputfile outputfile                (print to screen  )
c       % dsopt inputfile outputfile  >  listfile   (print to listfile)
c
c       If "inputfile"  is not specified, assumes "ds.dat"
c       If "outputfile" is not specified, assumes "ds.txt"
c
c
c     Output:
c      - DS speed and parameter printout for each optimization step
c      - outputfile  Time-trace file for the final orbit
c
c      - layer1.txt  file with top shear layer boundary x,y,z points
c      - layer2.txt  file with bot shear layer boundary x,y,z points 
c
c
c     Mark Drela
c     MIT Aero & Astro
c     29 Sep 2008
c-------------------------------------------------------------------------
      implicit real*8 (a-h,m,o-z)
c
      parameter (idim=361)
      dimension
     &     t(idim), 
     &     x(idim), 
     &     y(idim),
     &     z(idim),
     &    xt(idim), 
     &    yt(idim),
     &    zt(idim),
     &    xn(idim), 
     &    yn(idim),
     &    zn(idim),
     &    ux(idim), 
     &    uy(idim), 
     &    uz(idim), 
     &    us(idim),
     & windx(idim),
     & windy(idim),
     & windz(idim),
     &  vair(idim),
     &    cl(idim),
     &    cd(idim),
     &   cdw(idim),
     &  gfac(idim),
     &  time(idim),
     & omega(idim),
     &  roll(idim)
c
      character*80 argp1, argp2, fname
      character*128 line
c
      parameter (ldim=8)
      dimension par(ldim), pardel(ldim), dpar(ldim)
      dimension fob_par(ldim)
      dimension lparopt(0:ldim)
c
cc    data eps  / 1.0e-5 /    ! time step convergence tolerance
cc    data veps / 1.0e-4 /    ! orbit fractional velocity convergence tolerance
c
      data eps  / 1.0d-8 /    ! time step convergence tolerance
      data veps / 1.0d-6 /    ! orbit fractional velocity convergence tolerance
c
      data itmax  / 20 /   ! max number of Newton iterations per orbit interval
      data norbit / 20 /   ! max number of orbit iterations to converge speed
c
      version = 1.14
c
      call getarg(1,argp1)
      call getarg(2,argp2)
c
c
c---- design-parameter indices
      lmass = 1
      lrorb = 2
      laorb = 3
      lalfa = 4
      lphi  = 5
      lxlay = 6
      lzlay = 7
      npar = 7
c
c
c=================================================================================
c     Input section
c
      fname = argp1
      if(fname.eq.' ') fname = 'ds.dat'
c
      lu = 1
      open(lu,file=fname,status='old')
c
      iline = 0
c
      call rget(lu,iline, gee)
      call rget(lu,iline, rho)
      call rget(lu,iline, vsound)
c
      call rget(lu,iline, cd0)
      call rget(lu,iline, cd1)
      call rget(lu,iline, cd2)
      call rget(lu,iline, cd8)

      call rget(lu,iline, mcrit0)
      call rget(lu,iline, dmcritcl)
      call rget(lu,iline, cdwcon)
c
      call rget(lu,iline, amass)
      call rget(lu,iline, span)
      call rget(lu,iline, ar)
      call rget(lu,iline, speff)
c
      call rget(lu,iline, rorb)
      call rget(lu,iline, aorb)
      call rget(lu,iline, alfa)
      call rget(lu,iline, phi )
c
      call rget(lu,iline, wind1)
      call rget(lu,iline, wind2)
c
      call rget(lu,iline, xlayer)
      call rget(lu,iline, zlayer)
      call rget(lu,iline, txcon)
c
      call iget(lu,iline, norbit)
      call rget(lu,iline, vel0)
      call rget(lu,iline, cl0)
c
      call iget(lu,iline, nint)
      call iget(lu,iline, intv)
c
c
      call iget(lu,iline, nopt)
      call rget(lu,iline, optstep)

      call iget(lu,iline, lparopt(lmass))
      call iget(lu,iline, lparopt(lrorb))
      call iget(lu,iline, lparopt(laorb))
      call iget(lu,iline, lparopt(lalfa))
      call iget(lu,iline, lparopt(lphi)) 
      call iget(lu,iline, lparopt(lxlay))
      call iget(lu,iline, lparopt(lzlay))
c
      close(lu)
c
c=================================================================================
c---- number of orbit points, clipped to array limit
      n = min( nint+1 , idim )
      ivel = intv + 1
c
c-------------------------------------------------------------------
c---- derived quantities
c
c---- airplane reference area
      area = span**2 / ar
c
c---- shear layer spreading rate  d(thickness)/dx
      tlayerx = 2.0*txcon * (wind1-wind2)/(wind1+wind2)
c
c-------------------------------------------------------------------
c
c---- store optimized parameters in convenient array
      par(lmass) = amass
      par(lrorb) = rorb
      par(laorb) = aorb
      par(lalfa) = alfa
      par(lphi)  = phi
      par(lxlay) = xlayer
      par(lzlay) = zlayer
c
c---- perturbation changes for computing design gradients
      dpar(lmass) = amass * 0.015
      dpar(lrorb) = rorb * 0.015
      dpar(laorb) = 0.01
      dpar(lalfa) = 0.5
      dpar(lphi)  = -0.5
      dpar(lxlay) = rorb * 0.020
      dpar(lzlay) = rorb * 0.005
c
c---- always do baseline solution
      lparopt(0) = 1
c
      do 800 iopt = 0, nopt
      do 700 ipar = 0, npar
c
c------ skip parameter-sensitivity calculation if this parameter is not tweaked
        if(lparopt(ipar) .eq. 0) go to 700
c
c------ perturb parameter # ipar
        do jpar = 1, npar
          pardel(jpar) = par(jpar)
        enddo
        if(ipar.gt.0) pardel(ipar) = par(ipar) + dpar(ipar)
c
c------ set parameters for passing to routines
        amass  = pardel(lmass)
        rorb   = pardel(lrorb)
        aorb   = pardel(laorb)
        alfa   = pardel(lalfa)
        phi    = pardel(lphi) 
        xlayer = pardel(lxlay)
        zlayer = pardel(lzlay)
c
c------ set path geometry
        call setpath(n,alfa,phi,rorb,aorb,
     &             t, x, y, z,
     &                xt,yt,zt,
     &                xn,yn,zn )
c
c------ set wind velocity for each point in orbit
        do i = 1, n
          call windfield(wind1, wind2, 
     &                   tlayerx, xlayer, zlayer,
     &                   x(i), y(i), z(i), 
     &                   windx(i), windy(i), windz(i) )
        enddo
c
c------ compute converged orbit
        iprint = 0
        call orbit(norbit, veps, vel0, cl0, iprint,
     &           n,rho,vsound,gee,
     &           amass,area,ar,speff,
     &           cd0,cd1,cd2,cd8,
     &           mcrit0,dmcritcl,cdwcon,
     &           itmax,eps,
     &           x , y , z ,
     &           xt, yt, zt,
     &           xn, yn, zn,
     &           ux, uy, uz, us,
     &           windx, windy, windz,
     &           vair, cl, cd, cdw, gfac, time )
c
c------ if no optimization is to be done, jump out of parameter-perturbation loop
        if(nopt.eq.0) go to 710
c
        if(ipar.eq.0) then
c------- set baseline objective function (-DS speed)
         fob0 = -us(ivel)
        else
c------- set objective function and parameter derivatives
         fob = -us(ivel)
         fob_par(ipar) = (fob-fob0) / dpar(ipar) 
        endif
c
 700  continue
c
c---- modify parameters opposite objective-function gradient
      do ipar = 1, npar
        if(lparopt(ipar) .ne. 0) then
         par(ipar) = par(ipar) - fob_par(ipar)*dpar(ipar)**2 * optstep
        endif
      enddo
c
 710  continue
c
      if(iopt.eq.0) then
       write(*,1000)
     & '#    Speed     mass     R_orb    AR_orb',
     & '  alfa_orb  phi_orb  Xlayer   Zlayer'
       write(*,1000)
     & '#    [mph]     [kg]      [m]           ',
     & '    [deg]    [deg]     [m]      [m]'
      endif
 1000 format(a,a,a,a)
c
      speed = us(ivel)
      speedmph = speed * 3.6/1.609
      write(*,2050) 
     & speedmph,
     & par(lmass),
     & par(lrorb),
     & par(laorb),
     & par(lalfa),
     & par(lphi) ,
     & par(lxlay),
     & par(lzlay)
 2050 format(1x, f10.4, f9.4, f10.3, f9.4, 2f9.3, f9.3, f9.3)
c
 800  continue
c
c========================================================================
c---- post-processing and output
c
c---- calculate zero-moment roll angles
      call setroll(n, span,
     &             wind1, wind2, 
     &             tlayerx, xlayer, zlayer,
     &           x , y , z ,
     &           xt, yt, zt,
     &           xn, yn, zn,
     &           ux, uy, uz,
     &           windx, windy, windz,
     &           time, omega, roll )
c
c
c----------------------------------------------------------
      fname = argp2
      if(fname.eq.' ') fname = 'ds.txt'
c
      lu = 10
      open(lu,file=fname,status='unknown')
      rewind(lu)
      call wrtpath(lu, n,
     &        x, y, z,
     &        xt,yt,zt,
     &        ux,uy,uz, us,
     &        windx,windy,windz,
     &        vair,cl,cd,cdw,gfac,time, roll,vsound )
      close(lu)
c
c- - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
c---- optional: write out shear layer boundary surfaces for overplotting on orbit
      lu1 = 11
      lu2 = 12
      open(lu1,file='layer1.txt',status='unknown')
      open(lu2,file='layer2.txt',status='unknown')
      rewind(lu1)
      rewind(lu2)
      call wrtlayer(lu1,lu2,tlayerx, xlayer, zlayer, n,x,y,z)
      close(lu1)
      close(lu2)
c
      stop
      end ! dsopt



      subroutine rget(lu,iline, rvar)
      implicit real*8 (a-h,m,o-z)
      character*256 line
c
 10   continue
      iline = iline + 1
c
      read(lu,1000,end=90,err=90) line
 1000 format(a)
      if(index('!#',line(1:1)).ne.0) go to 10
      if(line .eq.' ') go to 10
c
      llen = index(line,'!') - 1
      if(llen .le. 0) llen = len(line)
c
      read(line(1:llen),*,err=90) rvar
      return
c
 90   continue
      write(*,*) 'File read error on line #', iline
      stop
      end ! rget



      subroutine iget(lu,iline, ivar)
      implicit real*8 (a-h,m,o-z)
      character*256 line
c
 10   continue
      iline = iline + 1
c
      read(lu,1000,end=90,err=90) line
 1000 format(a)
      if(index('!#',line(1:1)).ne.0) go to 10
      if(line.eq.' ') go to 10
c
      llen = index(line,'!') - 1
      if(llen .le. 0) llen = len(line)
c
      read(line(1:llen),*,err=90) ivar
      return
c
 90   continue
      write(*,*) 'File read error on line #', iline
      stop
      end ! iget


      subroutine setpath(n,alfa,phi,rorb,aorb,
     &                   t, x, y, z,
     &                      xt,yt,zt,
     &                      xn,yn,zn )
c---------------------------------------------------------------
c     Sets coordinates x,y,z of an ellipse in space.
c     Sets coordinate parametric derivatives  xt,yt,zt.
c     Sets normal direction vector components xn,yn,zn.
c---------------------------------------------------------------
      implicit real*8 (a-h,m,o-z)
      dimension
     &     t(*), 
     &     x(*), 
     &     y(*),
     &     z(*),
     &    xt(*), 
     &    yt(*),
     &    zt(*),
     &    xn(*), 
     &    yn(*),
     &    zn(*)
c
      pi = 4.0*atan(1.0)
c
      sina = sin(alfa*pi/180.0)
      cosa = cos(alfa*pi/180.0)
c
      sinp = sin(phi*pi/180.0)
      cosp = cos(phi*pi/180.0)
c
c---- ellipse major, minor semi-axes
      a = rorb
      b = rorb / aorb
c
c---- set orbit path coordinates
      do i = 1, n
c------ set path parameter = circle angle
        frac = float(i-1) / float(n-1)
        t(i) = 2.0*pi*frac
c
        cost = cos(t(i))
        sint = sin(t(i))
c
c------ 2D ellipse coordinates and t-derivatives
        xe = a*cost
        ze = b*sint
c
        xet = -a*sint
        zet =  b*cost
c
        xen = -zet / sqrt(xet**2 + zet**2)
        zen =  xet / sqrt(xet**2 + zet**2)
c
c------ rotate by tilt angle phi
        x0  = xe
        y0  = ze *sinp
        z0  = ze *cosp
c
        x0t = xet
        y0t = zet*sinp
        z0t = zet*cosp
c
        x0n = xen
        y0n = zen*sinp
        z0n = zen*cosp
c
c------ rotate by inclination angle alpha
        x(i)  = x0 *cosa - z0 *sina
        y(i)  = y0
        z(i)  = x0 *sina + z0 *cosa
c
        xt(i) = x0t*cosa - z0t*sina
        yt(i) = y0t
        zt(i) = x0t*sina + z0t*cosa
c
        xn(i) = x0n*cosa - z0n*sina
        yn(i) = y0n
        zn(i) = x0n*sina + z0n*cosa
      enddo
c
      return
      end ! setpath



      subroutine windfield(wind1, wind2, 
     &                     tlayerx, xlayer, zlayer,
     &                     x , y , z , 
     &                     windx, windy, windz )
      implicit real*8 (a-h,m,o-z)
c
      pi = 4.0*atan(1.0)
c
      xdel = x - xlayer
      zdel = z - zlayer
c
      tlayer = -xdel*tlayerx
      tlayer = max( tlayer , 1.0e-5 )
c     
      znorm = zdel / tlayer
      znorm = max( -0.5 , min( 0.5 , znorm ) )
c
      frac = 0.5 + 0.5*sin(pi*znorm)
      windx = -wind1*frac - wind2*(1.0-frac) 
      windy = 0.
      windz = 0.
c
      return
      end ! windfield



      subroutine wrtpath(lu, n,
     &        x, y, z,
     &        xt,yt,zt,
     &        ux,uy,uz, us,
     &        windx,windy,windz,
     &        vair,cl,cd,cdw,gfac,time , roll,vsound )
c
      implicit real*8 (a-h,m,o-z)
      dimension
     &     x(*),
     &     y(*),
     &     z(*),
     &    xt(*),
     &    yt(*),
     &    zt(*),
     &    ux(*), 
     &    uy(*), 
     &    uz(*), 
     &    us(*),
     & windx(*),
     & windy(*),
     & windz(*),
     &  vair(*),
     &    cl(*),
     &    cd(*),
     &   cdw(*),
     &  gfac(*),
     &  time(*),
     &  roll(*)
c
 1000 format(a,a,a,a)
c
      write(lu,1000)
     &     '#    1        2       3       4       5     ',
     &  '   6        7        8       9      10    ',
     &  '   11       12     13       14     15 '
      write(lu,1000)
     &     '#  frac       t       x       y       z     ',
     &  '   V        V      Vair    Mach   cdw/cd  ',
     &  '  Gload    CL     L/D    Vwind    Roll '
      write(lu,1000)
     &     '#            [s]     [m]     [m]     [m]    ',
     &  ' [m/s]    [mph]    [mph]                  ',
     &  '                         [mph]   [deg]'
c
      do i = 1, n
        frac = float(i-1) / float(n-1)
        vmph = us(i) * 3600.0/1609.0
        vairmph = vair(i) * 3600.0/1609.0
        mach = vair(i)/vsound
        cdwfrac = cdw(i)/cd(i)
        alod = cl(i)/cd(i)
c
        vwind = sqrt(windx(i)**2 + windy(i)**2 + windz(i)**2)
        vwindmph = vwind * 3600.0/1609.0
c
        write(lu,2100)
     &     frac, time(i), x(i), y(i), z(i), 
     &     us(i), vmph, vairmph,  mach, cdwfrac,
     &     gfac(i), cl(i), alod, vwindmph, roll(i)
c
 2100   format(1x,f8.4, f9.5, 3f8.3, 
     &     3f9.3, f7.3, f8.4,
     &     f9.2,  f8.4, f7.2, f9.3, f7.2)
      enddo
c
      return
      end ! wrtpath


      subroutine wrtlayer(lu1,lu2,tlayerx,xlayer,zlayer, n,x,y,z)
      implicit real*8 (a-h,m,o-z)
c
      dimension x(*), y(*), z(*)
c
      xmin = x(1)
      xmax = x(1)
      ymin = y(1)
      ymax = y(1)
      do i = 2, n
        xmin = min(x(i),xmin)
        xmax = max(x(i),xmax)
        ymin = min(y(i),ymin)
        ymax = max(y(i),ymax)
      enddo
c
      delx = (xmax - xmin)*1.1
      dely = (ymax - ymin)*1.1
      delx = max( delx , 0.75*dely )
      dely = max( dely , 0.75*delx )
c
      x1 = 0.5*(xmax+xmin) - 0.5*delx
      x2 = 0.5*(xmax+xmin) + 0.5*delx
      y1 = 0.5*(ymax+ymin) - 0.5*dely
      y2 = 0.5*(ymax+ymin) + 0.5*dely
c
      t1 = -(x1-xlayer)*tlayerx
      t2 = -(x2-xlayer)*tlayerx
c
      if(t1 .lt. 0.0 .and. tlayerx .gt. 0.0) then
       x1 = xlayer
       t1 = 0.
      endif
c
      if(t2 .lt. 0.0 .and. tlayerx .gt. 0.0) then
       x2 = xlayer
       t2 = 0.
      endif
c
      nx = 21
      ny = 21
c
c
      do iy = 1, ny
        fracy = float(iy-1)/float(ny-1)
        yy = y1*(1.0-fracy) + y2*fracy
c
        do ix = 1, nx
          fracx = float(ix-1)/float(nx-1)
          xx = x1*(1.0-fracx) + x2*fracx
          tt = t1*(1.0-fracx) + t2*fracx
          write(lu1,*) xx, yy, zlayer + 0.5*tt
          write(lu2,*) xx, yy, zlayer - 0.5*tt
        enddo
        write(lu1,*) ' '
        write(lu2,*) ' '
      enddo
c
      return
      end ! wrtlayer


      subroutine setroll(n, span,
     &             wind1, wind2, 
     &             tlayerx, xlayer, zlayer,
     &           x , y , z ,
     &           xt, yt, zt,
     &           xn, yn, zn,
     &           ux, uy, uz,
     &           windx, windy, windz,
     &           time, omega, roll)
      implicit real*8 (a-h,m,o-z)
      dimension
     &     x(*),
     &     y(*),
     &     z(*),
     &    xt(*),
     &    yt(*),
     &    zt(*),
     &    xn(*), 
     &    yn(*),
     &    zn(*),
     &    ux(*), 
     &    uy(*), 
     &    uz(*), 
     & windx(*),
     & windy(*),
     & windz(*),
     &  time(*),
     & omega(*),
     &  roll(*)
c
      pi = 4.0*atan(1.0)
c
c---- set wind velocities at each wind tip, for computing roll rates
      do i = 1, n
c------ set spanwise unit vector = rt x rn
        ex = yt(i)*zn(i) - zt(i)*yn(i)
        ey = zt(i)*xn(i) - xt(i)*zn(i)
        ez = xt(i)*yn(i) - yt(i)*xn(i)
        es = sqrt(ex**2 + ey**2 + ez**2)
        ex = ex/es
        ey = ey/es
        ez = ez/es
c
c------ effective semispan where roll moment AoA effectively acts
        rspan = 0.5*span * 0.8
c
c------ wind at right tip
        xtip = x(i) + rspan*ex
        ytip = y(i) + rspan*ey
        ztip = z(i) + rspan*ez
        call windfield(wind1, wind2, 
     &                 tlayerx, xlayer, zlayer,
     &                 xtip, ytip, ztip, 
     &                 wxr, wyr, wzr )
c
c------ wind at left tip
        xtip = x(i) - rspan*ex
        ytip = y(i) - rspan*ey
        ztip = z(i) - rspan*ez
        call windfield(wind1, wind2, 
     &                 tlayerx, xlayer, zlayer,
     &                 xtip, ytip, ztip, 
     &                 wxl, wyl, wzl )
c
c------ relative air velocity seen by centerline, tips
        vx = windx(i) - ux(i)
        vy = windy(i) - uy(i)
        vz = windz(i) - uz(i)
c
        vxr = wxr - ux(i)
        vyr = wyr - uy(i)
        vzr = wzr - uz(i)
c
        vxl = wxl - ux(i)
        vyl = wyl - uy(i)
        vzl = wzl - uz(i)
c
c------ set vector wx,wy,wz normal to center air velocity, along lift direction
        vsq = vx**2 + vy**2 + vz**2
        vdn = xn(i)*vx + yn(i)*vy + zn(i)*vz
        wx = xn(i) - vdn*vx/vsq
        wy = yn(i) - vdn*vy/vsq
        wz = zn(i) - vdn*vz/vsq
        ws = sqrt(wx**2 + wy**2 + wz**2)
c
        wx = wx/ws
        wy = wy/ws
        wz = wz/ws
c
c------ tip velocities along wx,wy,wz
        vnl = vxl*wx + vyl*wy + vzl*wz
        vnr = vxr*wx + vyr*wy + vzr*wz
c
c------ zero-moment rotation rate
        omega(i) = (vnr-vnl) / (2.0*rspan)
c
c------ normalized roll rate "phat"
c        vs = sqrt(vsq)
c        phat = omega(i)*span / (2.0*vs)
c
c------ integrate roll angle
        if(i.eq.1) then
         roll(i) = 0.
        else
         roll(i) = roll(i-1)
     &           + 0.5*(omega(i)+omega(i-1))
     &                *( time(i)- time(i-1)) * 180.0/pi
        endif
      enddo
c
      return
      end ! setroll

