C***********************************************************************
C    Module:  orbit.f
C 
C    Copyright (C) 2008 Mark Drela 
C 
C    This program is free software; you can redistribute it and/or modify
C    it under the terms of the GNU General Public License as published by
C    the Free Software Foundation; either version 2 of the License, or
C    (at your option) any later version.
C
C    This program is distributed in the hope that it will be useful,
C    but WITHOUT ANY WARRANTY; without even the implied warranty of
C    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
C    GNU General Public License for more details.
C
C    You should have received a copy of the GNU General Public License
C    along with this program; if not, write to the Free Software
C    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
C***********************************************************************

      subroutine orbit(norbit, veps, vel0,cl0,iprint,
     &           n,rho,vsound,gee,
     &           amass,area,ar,speff,
     &           cd0,cd1,cd2,cd8,
     &           mcrit0,dmcritcl,cdwcon,
     &           itmax,eps,
     &           x , y , z ,
     &           xt, yt, zt,
     &           xn, yn, zn,
     &           ux, uy, uz, us,
     &           windx, windy, windz,
     &           vair, cl, cd, cdw, gfac, time )
c---------------------------------------------------------------------------
c     Inverse-dynamics solver to compute an equilibrium DS orbit 
c     having a prescribed shape.
c
c     - Calls subroutine ORBIT1 which does all the dynamics calculations.
c     - Adjusts the initial speed given to ORBIT1 so as to
c        reach equilibrium, with final speed = initial speed.
c
c     Input:
c     ------
c       norbit  max number of orbits computed to reach equilibrium
c       veps    equilibrium convergence tolerance (on fractional speed change)
c       vel0    guess of initial speed for first orbit  (value not critical)
c       cl0     guess of initial CL for first orbit     (value not critical)
c       iprint   0 = don't print convergence info
c                1 = print convergence info
c       n       number of points in orbit
c       rho     air density
c       vsound  air speed of sound
c       gee     gravity
c       amass   airplane amass
c       area    airplane reference area
c       ar      airplane aspect ratio
c       speff   induced drag span efficiency
c       cd0     constants for profile drag cdp(cl) function
c       cd1      "
c       cd2      "
c       cd8      "
c       mcrit0   constants for wave drag cdwave(M,cl) function
c       dmcritcl  "
c       cdwcon    "
c       itmax   max number of Newton iterations to converge each interval
c       eps     Newton convergence tolerance
c
c       x ,y ,z(i)   all path coordinates
c       xt,yt,zt(i)  vector along path
c       xn,yn,zn(i)  vector normal to path
c       windx,windy,windz(i)   wind velocity components
c
c     Output:
c     -------
c       ux,uy,uz(i)   airplane velocity components
c       us(i)         airplane groundspeed  =  sqrt(ux^2 + uy^2 + uz^2)
c       vair(i)       airplane airspeed
c       cl(i)         CL
c       cd(i)         CD
c       cdw(i)        CDwave
c       gfac(i)       G-load factor
c       time(i)       time
c
c---------------------------------------------------------------------------
      implicit real*8 (a-h,m,o-z)
c
      dimension
     &     x(*), 
     &     y(*),
     &     z(*),
     &    xt(*), 
     &    yt(*),
     &    zt(*),
     &    xn(*), 
     &    yn(*),
     &    zn(*),
     &    ux(*), 
     &    uy(*), 
     &    uz(*), 
     &    us(*),
     & windx(*),
     & windy(*),
     & windz(*),
     &  vair(*),
     &    cl(*),
     &    cd(*),
     &   cdw(*),
     &  gfac(*),
     &  time(*)
c
c
      do 500 iorbit = 1, norbit
c
c------ set values for initial point
        i = 1
        if(iorbit.eq.1) then
c------- first orbit... use initial guesses
         time(i) = 0.
         us(i) = vel0
         cl(i) = cl0
         st = sqrt(xt(i)**2 + yt(i)**2 + zt(i)**2)
         ux(i) = vel0 * xt(i)/st
         uy(i) = vel0 * yt(i)/st
         uz(i) = vel0 * zt(i)/st
c
        else
c------- subsequent orbits... use previous orbit's last point plus Newton correction
ccc      time(i) = time(n)
         time(i) = 0.
c
c------- Newton residual to drive 1st and last orbit speeds to be the same
         res = us(n) - us(i)
         res_u0 = usn_u0 - 1.0
         du0 = -res / res_u0
c
         rlx = 1.0
         if(du0 .gt.  0.05*us(i)) rlx =  0.05*us(i)/du0
         if(du0 .lt. -0.05*us(i)) rlx = -0.05*us(i)/du0
c
c------- Newton correction to initial orbit speed
         us(i) = us(i) + rlx*du0
c
         st = sqrt(xt(i)**2 + yt(i)**2 + zt(i)**2)
         ux(i) = us(i)*xt(i)/st
         uy(i) = us(i)*yt(i)/st
         uz(i) = us(i)*zt(i)/st
c
c------- also set output values for i=1, since these aren't set by orbit1
         cl(i)   = cl(n)
         cd(i)   = cd(n)
         vair(i) = vair(n)
         gfac(i) = gfac(n)
c
        endif
c
        call orbit1(n,rho,vsound,gee,
     &            amass,area,ar,speff,
     &            cd0,cd1,cd2,cd8,
     &            mcrit0,dmcritcl,cdwcon,
     &            itmax,eps,
     &            usn_u0,
     &            x , y , z ,
     &            xt, yt, zt,
     &            xn, yn, zn,
     &            ux, uy, uz, us,
     &            windx, windy, windz,
     &            vair, cl, cd, cdw, gfac, time )
c
c------ check velocity convergence
        delus = us(n) - us(1)
c
        if(iprint.ne.0) then
         write(*,1500) iorbit, delus/us(n)
 1500    format(1x,'Orbit',i3,':  dV/V =', e11.3)
        endif
c
c------ return if converged
        if(abs(delus) .lt. veps*us(n)) return
c
 500  continue ! with next orbit
      write(*,*) 'ORBIT: Equilibrium not reached.'
c
      return
      end ! orbit


      subroutine orbit1(n,rho,vsound,gee,
     &            amass,area,ar,speff,
     &            cd0,cd1,cd2,cd8,
     &            mcrit0,dmcritcl,cdwcon,
     &            itmax,eps, 
     &            usn_u0,
     &            x , y , z ,
     &            xt, yt, zt,
     &            xn, yn, zn,
     &            ux, uy, uz, us,
     &            windx, windy, windz,
     &            vair, cl, cd, cdw, gfac, time )
c----------------------------------------------------------------------
c     Computes one DS orbit of prescribed shape.
c
c     Input:
c     ------
c       n         same as for ORBIT -- see header above
c       .           "
c       .           "
c       windz(i)    "
c
c       us(1)     initial speed
c       cl(1)     initial cl guess (value not critical)
c
c       
c     Output:
c     -------
c       ux(i)     same as for ORBIT -- see header above
c       .           "
c       .           "
c       time(i)     "
c
c       usn_u0    sensitivity   d us(n) / d us(1)
c                 ( useful to converge initial speed towards orbit equilibrium )
c---------------------------------------------------------------------
      implicit real*8 (a-h,m,o-z)
      dimension
     &     x(*), 
     &     y(*),
     &     z(*),
     &    xt(*),
     &    yt(*),
     &    zt(*),
     &    xn(*),
     &    yn(*),
     &    zn(*),
     &    ux(*), 
     &    uy(*), 
     &    uz(*), 
     &    us(*),
     & windx(*),
     & windy(*),
     & windz(*),
     &  vair(*),
     &    cl(*),
     &    cd(*),
     &   cdw(*),
     &  gfac(*),
     &  time(*)
c
      pi = 4.0*atan(1.0)
c
      i = 1
      st = sqrt(xt(i)**2 + yt(i)**2 + zt(i)**2)
      usm_u0 = 1.0
      uxm_u0 = xt(i)/st
      uym_u0 = yt(i)/st
      uzm_u0 = zt(i)/st
c
c---- go over i-1..i orbit path intervals
      do 100 i = 2, n
        dx = x(i)-x(i-1)
        dy = y(i)-y(i-1)
        dz = z(i)-z(i-1)
c
        ds = sqrt(dx**2 + dy**2 + dz**2)
c
c------ unit vector along path interval
        xs = dx/ds
        ys = dy/ds
        zs = dz/ds
c
c------ vector normal to path interval (magnitude not significant here)
        xna = 0.5*(xn(i)+xn(i-1))
        yna = 0.5*(yn(i)+yn(i-1))
        zna = 0.5*(zn(i)+zn(i-1))
c
c------ initial guesses for speed and CL, using previous point
        us(i) = us(i-1)
        cl(i) = cl(i-1)
c
c------ Newton iteration to converge on exact speed and CL
        do 50 iter = 1, itmax
c
c-------- average speed over interval
          usa = 0.5*(us(i)+us(i-1))
          usa_us = 0.5
c
c-------- time step over interval
          dtime = ds / usa
          dtime_us = -dtime/usa * usa_us
c
c-------- velocity vector components
          st = sqrt(xt(i)**2 + yt(i)**2 + zt(i)**2)
          ux1 = xt(i)/st
          uy1 = yt(i)/st
          uz1 = zt(i)/st
c
          ux(i) = us(i) * ux1
          uy(i) = us(i) * uy1
          uz(i) = us(i) * uz1
c
c-------- accelerations
          ax = (ux(i) - ux(i-1))/dtime
          ay = (uy(i) - uy(i-1))/dtime
          az = (uz(i) - uz(i-1))/dtime
c
          ax_us = ux1/dtime - ax/dtime * dtime_us
          ay_us = uy1/dtime - ay/dtime * dtime_us
          az_us = uz1/dtime - az/dtime * dtime_us
c
c-------- reset az to be relative to inertial space
          az = az + gee
c
c-------- air velocity components seen by plane
          vx = 0.5*(windx(i)+windx(i-1)) - 0.5*(ux(i)+ux(i-1))
          vy = 0.5*(windy(i)+windy(i-1)) - 0.5*(uy(i)+uy(i-1))
          vz = 0.5*(windz(i)+windz(i-1)) - 0.5*(uz(i)+uz(i-1))
c
          vx_us = -0.5*ux1
          vy_us = -0.5*uy1
          vz_us = -0.5*uz1
c
          vsq = vx**2 + vy**2 + vz**2
          va = sqrt(vsq)
c
          vsq_us = 2.0*(vx*vx_us + vy*vy_us + vz*vz_us)
          va_us  = 0.5*vsq_us/va
c
          mach = va/vsound
          mach_us = va_us/vsound
c
c-------- dynamic pressure
          q = 0.5*rho*vsq
          q_us = 0.5*rho*vsq_us
c
c-------- drag coefficients
          cdi  = cl(i)*cl(i) / (pi*ar*speff)
          cdi_cl = 2.0*cl(i) / (pi*ar*speff)
c
          mcrit = mcrit0 + dmcritcl*cl(i)
          mcrit_cl =       dmcritcl
          if(mach .le. mcrit) then
           cdw(i) = 0.
           cdw_us = 0.
           cdw_cl = 0.
          else
           cdw(i) =     cdwcon*(mach-mcrit)**4
           cdw_us = 4.0*cdwcon*(mach-mcrit)**3 * mach_us
           cdw_cl = 4.0*cdwcon*(mach-mcrit)**3 * (-mcrit_cl)
          endif

          cd(i) = cdi
     &          + cd0
     &          + cd1*cl(i)
     &          + cd2*cl(i)**2
     &          + cd8*cl(i)**8
     &          + cdw(i)
          cd_cl = cdi_cl
     &          + cd1
     &          + cd2*cl(i)    * 2.0
     &          + cd8*cl(i)**7 * 8.0
     &          + cdw_cl
          cd_us = cdw_us
c
c-------- aero lift and drag
          flift    = q   *area*cl(i)
          flift_cl = q   *area
          flift_us = q_us*area*cl(i)
c
          fdrag    = q   *area*cd(i)
          fdrag_cl = q   *area*cd_cl
          fdrag_us = q   *area*cd_us
     &             + q_us*area*cd(i)
c
c-------- set vector wx,wy,wz normal to air velocity, and in plane of orbit
          vdn    = xna*vx    + yna*vy    + zna*vz
          vdn_us = xna*vx_us + yna*vy_us + zna*vz_us
c
          wx = xna - vdn*vx/vsq
          wy = yna - vdn*vy/vsq
          wz = zna - vdn*vz/vsq
c
          wx_us = -vdn_us*vx/vsq - vdn*vx_us/vsq + vdn*vx/vsq**2 *vsq_us
          wy_us = -vdn_us*vy/vsq - vdn*vy_us/vsq + vdn*vy/vsq**2 *vsq_us
          wz_us = -vdn_us*vz/vsq - vdn*vz_us/vsq + vdn*vz/vsq**2 *vsq_us
c
          wa = sqrt(wx**2 + wy**2 + wz**2)
          wa_us = ( wx*wx_us
     &            + wy*wy_us
     &            + wz*wz_us ) / wa
c
c-------- total aero force vector
          fx = flift*wx/wa + fdrag*vx/va
          fy = flift*wy/wa + fdrag*vy/va
          fz = flift*wz/wa + fdrag*vz/va
c
          fx_cl = flift_cl*wx/wa + fdrag_cl*vx/va
          fy_cl = flift_cl*wy/wa + fdrag_cl*vy/va
          fz_cl = flift_cl*wz/wa + fdrag_cl*vz/va
c
          fx_us = flift_us*wx/wa + fdrag_us*vx/va
          fy_us = flift_us*wy/wa + fdrag_us*vy/va
          fz_us = flift_us*wz/wa + fdrag_us*vz/va
c
          fx_us = flift*wx_us/wa + fdrag*vx_us/va + fx_us
          fy_us = flift*wy_us/wa + fdrag*vy_us/va + fy_us
          fz_us = flift*wz_us/wa + fdrag*vz_us/va + fz_us
c
          fx_us = - flift*(wx/wa**2)*wa_us 
     &            - fdrag*(vx/va**2)*va_us + fx_us
          fy_us = - flift*(wy/wa**2)*wa_us
     &            - fdrag*(vy/va**2)*va_us + fy_us
          fz_us = - flift*(wz/wa**2)*wa_us
     &            - fdrag*(vz/va**2)*va_us + fz_us
c
c
c-------- equation residual 1:  R1 = (f - ma).s = 0
          res1 = (fx - amass*ax)*xs
     &         + (fy - amass*ay)*ys
     &         + (fz - amass*az)*zs
          res1_us = 
     &           (fx_us - amass*ax_us)*xs
     &         + (fy_us - amass*ay_us)*ys
     &         + (fz_us - amass*az_us)*zs
          res1_cl = 
     &           (fx_cl             )*xs
     &         + (fy_cl             )*ys
     &         + (fz_cl             )*zs
c
c-------- equation residual 2:  R2 = (f - ma).n = 0
          res2 = (fx - amass*ax)*xna
     &         + (fy - amass*ay)*yna
     &         + (fz - amass*az)*zna
          res2_us = 
     &           (fx_us - amass*ax_us)*xna
     &         + (fy_us - amass*ay_us)*yna
     &         + (fz_us - amass*az_us)*zna
          res2_cl = 
     &           (fx_cl             )*xna
     &         + (fy_cl             )*yna
     &         + (fz_cl             )*zna
c
c
c-------- 2x2 Newton system matrix
          a11 = res1_us
          a12 = res1_cl
          a21 = res2_us
          a22 = res2_cl
c
          det =  a11*a22 - a12*a21
c
c-------- solve Newton system
          dus = -(res1*a22  - a12 *res2)/det
          dcl = -(a11 *res2 - res1*a21 )/det
c
c-------- find under-relaxation factor if needed
          rlx = 1.0
          if(rlx*dus .gt.  0.1*us(i)) rlx =  0.1*us(i)/dus
          if(rlx*dus .lt. -0.1*us(i)) rlx = -0.1*us(i)/dus
          if(rlx*dcl .gt.  0.1) rlx =  0.1/dcl
          if(rlx*dcl .lt. -0.1) rlx = -0.1/dcl
c
c      write(*,*) res1, res2
c      pause

c-------- Newton update
          cl(i) = cl(i) + rlx*dcl
          us(i) = us(i) + rlx*dus
c
c-------- exit Newton loop if changes are below convergence tolerance
          if(abs(dcl) .le. eps     .and. 
     &       abs(dus) .le. eps*us(i)     ) then
           go to 55
          endif
c
 50     continue ! with next Newton iteration
        write(*,*) 'ORBIT1: Newton iteration failed. dCL dV/V=',
     &              dcl, dus/us(i)
c
c------ pick up here after Newton iterations
 55     continue
c
c------ set final values
        ux(i) = us(i) * ux1
        uy(i) = us(i) * uy1
        uz(i) = us(i) * uz1
c
        vair(i) = va
c
        if(gee.gt.0.0) then
         gfac(i) = flift / (amass*gee)
        else
         gfac(i) = 0.
        endif
c
        usa = 0.5*(us(i)+us(i-1))
        dtime = ds / usa
        time(i) = time(i-1) + dtime
c
c------ calculate sensitivities  dus(i)/du0  and  dcl(i)/du0
c
          usa_u0 = 0.5*usm_u0
c
          dtime_u0 = -dtime/usa * usa_u0
c
          ax_u0 = -uxm_u0/dtime - ax/dtime * dtime_u0
          ay_u0 = -uym_u0/dtime - ay/dtime * dtime_u0
          az_u0 = -uzm_u0/dtime - az/dtime * dtime_u0
c
          vx_u0 = -0.5*uxm_u0
          vy_u0 = -0.5*uym_u0
          vz_u0 = -0.5*uzm_u0
c
          vsq_u0 = 2.0*(vx*vx_u0 + vy*vy_u0 + vz*vz_u0)
          va_u0  = 0.5*vsq_u0/va

          q_u0 = 0.5*rho*vsq_u0

          flift_u0 = q_u0*area*cl(i)
          fdrag_u0 = q_u0*area*cd(i)

          vdn_u0 = xna*vx_u0 + yna*vy_u0 + zna*vz_u0
          wx_u0 = -vdn_u0*vx/vsq - vdn*vx_u0/vsq + vdn*vx/vsq**2 *vsq_u0
          wy_u0 = -vdn_u0*vy/vsq - vdn*vy_u0/vsq + vdn*vy/vsq**2 *vsq_u0
          wz_u0 = -vdn_u0*vz/vsq - vdn*vz_u0/vsq + vdn*vz/vsq**2 *vsq_u0
c
          wa_u0 = ( wx*wx_u0
     &            + wy*wy_u0
     &            + wz*wz_u0 ) / wa

c-------- also set u0 derivatives for computing us_u0 after convergence
          fx_u0 = flift_u0*wx/wa + fdrag_u0*vx/va
          fy_u0 = flift_u0*wy/wa + fdrag_u0*vy/va
          fz_u0 = flift_u0*wz/wa + fdrag_u0*vz/va
c
          fx_u0 = flift*wx_u0/wa + fdrag*vx_u0/va + fx_u0
          fy_u0 = flift*wy_u0/wa + fdrag*vy_u0/va + fy_u0
          fz_u0 = flift*wz_u0/wa + fdrag*vz_u0/va + fz_u0
c
          fx_u0 = - flift*(wx/wa**2)*wa_u0 
     &            - fdrag*(vx/va**2)*va_u0 + fx_u0
          fy_u0 = - flift*(wy/wa**2)*wa_u0
     &            - fdrag*(vy/va**2)*va_u0 + fy_u0
          fz_u0 = - flift*(wz/wa**2)*wa_u0
     &            - fdrag*(vz/va**2)*va_u0 + fz_u0

c-------- also set u0 derivatives for computing us_u0 after convergence
          res1_u0 = 
     &           (fx_u0 - amass*ax_u0)*xs
     &         + (fy_u0 - amass*ay_u0)*ys
     &         + (fz_u0 - amass*az_u0)*zs
c
          res2_u0 = 
     &           (fx_u0 - amass*ax_u0)*xna
     &         + (fy_u0 - amass*ay_u0)*yna
     &         + (fz_u0 - amass*az_u0)*zna
c
        us_u0 = -(res1_u0*a22     - a12    *res2_u0)/det
        cl_u0 = -(a11    *res2_u0 - res1_u0*a21    )/det
c
c------ set these for next interval
        usm_u0 = us_u0
        uxm_u0 = us_u0*ux1
        uym_u0 = us_u0*uy1
        uzm_u0 = us_u0*uz1
c
 100  continue ! with next orbit interval
c
c---- set final value for returning
      usn_u0 = us_u0
c
      return
      end ! orbit1

